# -*- coding: future_fstrings -*-
from __future__ import unicode_literals

#
#  Copyright 2014, Jérôme Colin, Jérôme Kerdreux, Philippe Tanguy,
#  Telecom Bretagne.
#
#  This file is part of xAAL.
#
#  xAAL is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  xAAL is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with xAAL. If not, see <http://www.gnu.org/licenses/>.
#

from .network  import NetworkConnector
from .messages import MessageFactory,MessageType
from .exceptions  import *
from . import config

import time
import inspect
import collections
from enum import Enum

import logging
logger = logging.getLogger(__name__)

class EngineState(Enum):
    boot  = 0
    start = 1
    run   = 2
    halt  = 3
    dead  = 4

class Engine(object):

    def __init__(self,address=config.address,port=config.port,hops=config.hops,key=config.key):
        self.devices = []                        # list of devices / use (un)register_devices()
        self.state = EngineState.halt
        self.timers = []                         # functions to call periodic
        self.__last_timer = 0                    # last timer check
        self.subscribers =[self.handle_request]  # message receive workflow

        self.__attributesChange = []             # list of XAALAttributes instances
        self.__txFifo = collections.deque()      # tx msg fifo

        # start network
        self.network = NetworkConnector(address, port, hops)
        # start msg worker
        self.msg_factory = MessageFactory(key)

    #####################################################
    # Devices management
    #####################################################
    def add_device(self, dev):
        """register a new device """
        if dev not in self.devices:
            self.devices.append(dev)
            dev.engine = self

    def add_devices(self, devs):
        """register new devices"""
        for dev in devs:
            self.add_device(dev)

    def remove_device(self, dev):
        """unregister a device """
        dev.engine = None
        # Remove dev from devices list
        self.devices.remove(dev)

    #####################################################
    # xAAL messages Tx handling
    #####################################################
    # Fifo for msg to send
    def queue_msg(self, msg):
        """queue an encoded / cyphered message"""
        self.__txFifo.append(msg)

    def process_tx_msg(self):
        """ Process (send) message in tx queue called from the loop()"""
        cnt = 0
        while self.__txFifo:
            temp = self.__txFifo.popleft()
            self.send_msg(temp)
            # try to limit rate
            cnt = cnt + 1
            if cnt > config.queue_size:
                time.sleep(0.2)
                break

    def send_msg(self, msg):
        """Send an encoded message to the bus, use queue_msg instead"""
        self.network.send(msg)

    def send_request(self,dev,targets,action,body = None):
        """queue a new request"""
        msg = self.msg_factory.build_msg(dev, targets, MessageType.REQUEST, action, body)
        self.queue_msg(msg)

    def send_reply(self, dev, targets, action, body=None):
        """queue a new reply"""
        msg = self.msg_factory.build_msg(dev, targets, MessageType.REPLY, action, body)
        self.queue_msg(msg)

    def send_error(self, dev, errcode, description=None):
        """queue a error message"""
        msg = self.msg_factory.build_error_msg(dev, errcode, description)
        self.queue_msg(msg)

    def send_get_description(self,dev,targets):
        """queue a get_description request"""
        self.send_request(dev,targets,'get_description')

    def send_get_attributes(self,dev,targets):
        """queue a get_attributes request"""
        self.send_request(dev,targets,'get_attributes')

    def send_notification(self,dev,action,body=None):
        """queue a notificaton"""
        msg = self.msg_factory.build_msg(dev,[],MessageType.NOTIFY,action,body)
        self.queue_msg(msg)
        
    #####################################################
    # Alive messages
    #####################################################
    def send_alive(self, dev):
        """Send a Alive message for a given device"""
        timeout = dev.get_timeout()
        msg = self.msg_factory.build_alive_for(dev, timeout)
        self.queue_msg(msg)

    def send_is_alive(self, dev, dev_types="any.any"):
        """Send a is_alive message, w/ dev_types filtering"""
        body = {}
        body['dev_types'] = dev_types
        msg = self.msg_factory.build_msg(dev, [], MessageType.REQUEST, "is_alive", body)
        self.queue_msg(msg)

    def process_alives(self):
        """Periodic sending alive messages"""
        now = time.time()
        for dev in self.devices:
            if dev.next_alive < now :
                self.send_alive(dev)
                dev.update_alive()

    #####################################################
    # xAAL attributes changes
    #####################################################
    def add_attributes_change(self, attr):
        """add a new attribute change to the list"""
        self.__attributesChange.append(attr)

    def get_attributes_change(self):
        """return the pending attributes changes list"""
        return self.__attributesChange

    def process_attributes_change(self):
        """Processes (send notify) attributes changes for all devices"""
        devices = {}
        # Group attributes changed by device
        for attr in self.get_attributes_change():
            if attr.device not in devices.keys():
                devices[attr.device] = {}
            devices[attr.device][attr.name] = attr.value

        for dev in devices:
            self.send_notification(dev,"attributes_change",devices[dev])
        self.__attributesChange = []  # empty array

    #####################################################
    # xAAL messages subscribers
    #####################################################
    def receive_msg(self):
        """return new received message or None"""
        result = None
        data = self.network.get_data()
        if data:
            try:
                msg = self.msg_factory.decode_msg(data)
            except MessageParserError as e:
                logger.warning(e)
                msg = None
            result = msg
        return result

    def subscribe(self,func):
        self.subscribers.append(func)

    def unsubscribe(self,func):
        self.subscribers.remove(func)

    def process_subscribers(self):
        """process incomming messages"""
        msg = self.receive_msg()
        if msg:
            for func in self.subscribers:
                func(msg)
            self.process_attributes_change()
            
    def handle_request(self, msg):
        """Filter msg for devices according default xAAL API then process the
        request for each targets identied in the engine
        """
        if not msg.is_request():
            return 
        
        targets = filter_msg_for_devices(msg, self.devices)
        for target in targets:
            if msg.action == 'is_alive':
                self.send_alive(target)
            else:
                self.handle_action_request(msg, target)

    def handle_action_request(self, msg, target):
        """Run method (xAAL exposed method) on device:
            - None is returned if device method do not return anything
            - result is returned if device method gives a response
            - Errors are raised if an error occured:
                * Internal error
                * error returned on the xAAL bus
        """
        try:
            result = run_action(msg, target)
            if result != None:
                self.send_reply(dev=target,targets=[msg.source],action=msg.action,body=result)
        except CallbackError as e:
            self.send_error(target, e.code, e.description)
        except XAALError as e:
            logger.error(e)

    #####################################################
    # timers
    #####################################################
    def add_timer(self,func,period,counter=-1):
        """ 
        func: function to call
        period: period in second
        counter: number of repeat, -1 => always
        """
        if counter == 0:
            raise EngineError("Timer counter should =-1 or >0")
        t = Timer(func,period,counter)
        self.timers.append(t)
        return t

    def remove_timer(self,timer):
        """remove a given timer from the list"""
        self.timers.remove(timer)

    def process_timers(self):
        """Process all timers to find out which ones should be run"""
        expire_list = []
        
        if len(self.timers)!=0 :
            now = time.time()
            # little hack to avoid to check timer to often.
            # w/ this enable timer precision is bad, but far enougth
            if (now - self.__last_timer) < 0.4: return

            for t in self.timers:
                if t.deadline < now:
                    try:
                        t.func()
                    except CallbackError as e:
                        logger.error(e.description)
                    if t.counter != -1:
                        t.counter-= 1
                        if t.counter == 0:
                            expire_list.append(t)
                    t.deadline = now + t.period
            # delete expired timers
            for t in expire_list:
                self.remove_timer(t)
                    
            self.__last_timer = now

    #####################################################
    # Mainloops & run ..
    #####################################################
    def loop(self):
        """Process incomming xAAL msg
        Process timers
        Process attributes change for devices
        Process is_alive for device
        Send msgs from the Tx Buffer
        """
        # Process xAAL msg received, filter msg and process request
        self.process_subscribers()
        # Process timers
        self.process_timers()
        # Process attributes change for devices due to timers
        self.process_attributes_change()
        # Process Alives
        self.process_alives()
        # Process xAAL msgs to send
        self.process_tx_msg()

    def start(self):
        """Start the core engine: send queue alive msg"""
        if self.state in [EngineState.start,EngineState.run]:
            return
        self.network.connect()
        for dev in self.devices:
            self.send_alive(dev)
            dev.update_alive()
        self.state = EngineState.start

    def stop(self):
        self.state = EngineState.halt

    def shutdown(self):
        self.stop()

    def run(self):
        self.start()
        self.state = EngineState.run
        while self.state == EngineState.run:
            self.loop()


def filter_msg_for_devices(msg, devices):
    """loop throught the devices, to find which are
    expected w/ the msg

    - Filter on dev_types for is_alive request.
    - Filter on device address
    """
    results = []
    if msg.action == 'is_alive':
        if 'dev_types' in msg.body.keys():
            dev_types = msg.body['dev_types']
            if 'any.any' in dev_types:
                results = devices
            else:
                for dev in devices:
                    any_subtype = dev.dev_type.split('.')[0] + '.any'
                    if dev.dev_type in dev_types:
                        results.append(dev)
                    elif any_subtype in dev_types:
                        results.append(dev)
    else:
        if not msg.targets:  # if target list is empty == broadcast
            results = devices
        else:
            for dev in devices:
                if dev.address in msg.targets:
                    results.append(dev)
    return results

def run_action(msg,device):
    """ Extract an action & launch it
    Return:
        - action result
        - None if no result

    Note: If an exception raised, it's logged, and raise an XAALError.
    """
    method,params = search_action(msg,device)
    result = None
    try:
        result = method(**params)
    except Exception as e:
        logger.error(e)
        raise XAALError("Error in method:%s params:%s" % (msg.action,params))
    return result

def search_action(msg, device):
    """Extract an action (match with methods) from a msg on the device.
    Return:
        - None
        - found method & matching parameters

    Note: If method not found raise error, if wrong parameter error log
    """
    methods = device.get_methods()
    params = {}
    result = None
    if msg.action in methods.keys():
        method = methods[msg.action]
        body_params = None
        if msg.body:
            method_params = get_args_method(method)
            body_params = msg.body

            for k in body_params:
                temp = '_%s' %k
                if temp in method_params:
                    params.update({temp:body_params[k]})
                else:
                    logger.warning("Wrong method parameter [%s] for action %s" %(k, msg.action))
        result =  (method,params)
    else:
        raise XAALError("Method %s not found on device %s" % (msg.action,device))
    return result

def get_args_method(method):
    """ return the list on arguments for a given python method """
    spec = inspect.getargspec(method)
    try:
        spec.args.remove('self')
    except:
        pass
    return spec.args


class Timer(object):
    def __init__(self,func,period,counter):
        self.func = func
        self.period = period
        self.counter = counter
        self.deadline = time.time() + period
